"""This module gathers misc convenience functions to handle s3 objects"""
import re
import tempfile
from typing import Any, Dict, Optional, Tuple
from typing.io import BinaryIO
from urllib.parse import unquote, urlparse

import s3fs

S3_SCHEMES = ['s3', 's3n', 's3a']


def parse_s3_url(url: str, file=True) -> Tuple[Optional[str], Optional[str], Optional[str], str]:
    """parses a s3 url and extract credentials and s3 object path.

    A S3 URL looks like s3://aws_key:aws_secret@bucketname/objectname where
    credentials are optional. Since credentials might include characters
    such as `/`, `@` or `#`, they have to be urlquoted in the url.

    Args:
        url (str): the s3 url
        file (bool): whether or not the url is a file url or a directory one

    Returns:
        tuple: (access_key, secret, bucketname, objectname). If credentials
        are not specified, `access_key` and `secret` are set to None.
    """
    urlchunks = urlparse(url)
    scheme = urlchunks.scheme
    assert scheme in S3_SCHEMES, f'{scheme} unsupported, use one of {S3_SCHEMES}'
    assert not urlchunks.params, f's3 url should not have params, got {urlchunks.params}'
    assert not urlchunks.query, f's3 url should not have query, got {urlchunks.query}'
    assert not urlchunks.fragment, f's3 url should not have fragment, got {urlchunks.fragment}'

    access_key: Optional[str] = None
    secret: Optional[str] = None

    # if either username or password is specified, we have credentials
    if urlchunks.username or urlchunks.password:
        # and they should both not be empty
        assert urlchunks.username, 's3 access key should not be empty'
        assert urlchunks.password, 's3 secret should not be empty'
        access_key = unquote(urlchunks.username)
        secret = unquote(urlchunks.password)
    objectname = urlchunks.path.lstrip('/')  # remove leading /, it's not part of the objectname
    if file:
        assert objectname, "s3 objectname can't be empty"

    return access_key, secret, urlchunks.hostname, objectname


def s3_open(url: str, *, client_kwargs: Optional[Dict[str, Any]] = None) -> BinaryIO:
    """opens a s3 url and returns a file-like object"""
    access_key, secret, bucketname, objectname = parse_s3_url(url)
    fs = s3fs.S3FileSystem(key=access_key, secret=secret, client_kwargs=client_kwargs)
    ret = tempfile.NamedTemporaryFile(suffix='.s3tmp')
    file = fs.open(f'{bucketname}/{objectname}')
    ret.write(file.read())
    ret.seek(0)
    return ret


def _get_timestamp(obj: dict) -> Optional[int]:
    try:
        return obj['LastModified'].timestamp()
    except KeyError:
        return None


def s3_mtime(url: str, *, client_kwargs: Optional[Dict[str, Any]] = None) -> Optional[int]:
    access_key, secret, bucketname, objectname = parse_s3_url(url, file=True)
    fs = s3fs.S3FileSystem(key=access_key, secret=secret, client_kwargs=client_kwargs)
    return _get_timestamp(fs.info(f'{bucketname}/{objectname}'))


def dir_mtimes(
    dirpath: str, *, client_kwargs: Optional[Dict[str, Any]] = None
) -> Dict[str, Optional[int]]:
    access_key, secret, bucketname, objectname = parse_s3_url(dirpath, file=False)
    fs = s3fs.S3FileSystem(key=access_key, secret=secret, client_kwargs=client_kwargs)
    # objectname can be empty or not ('subdir1/subdir2')
    bucketdir = f'{bucketname}/{objectname}'.rstrip('/')
    return {
        re.sub(rf'^{bucketdir}/', '', x['name']): _get_timestamp(x) for x in fs.listdir(bucketdir)
    }



from abc import ABCMeta, abstractmethod
from contextlib import suppress
from datetime import timedelta
from enum import Enum
from pathlib import Path
from time import time
from typing import Optional, Union

import pandas as pd


class CacheEnum(str, Enum):
    MEMORY = 'memory'
    HDF = 'hdf'


class Cache(metaclass=ABCMeta):
    @staticmethod
    def get_cache(kind: CacheEnum, *args, **kwargs):
        return {CacheEnum.MEMORY: InMemoryCache, CacheEnum.HDF: HDFCache}[kind](*args, **kwargs)

    @staticmethod
    def should_invalidate(
        *,
        mtime: Optional[float] = None,
        cached_mtime: float,
        expire: Optional[timedelta] = None,
        cached_created_at: float,
    ):
        now = time()
        is_newer_version = False
        is_expired = False
        if mtime is not None:
            is_newer_version = mtime != cached_mtime
        if expire is not None:
            is_expired = now > cached_created_at + expire.total_seconds()
        return is_newer_version or is_expired

    @abstractmethod
    def get(
        self, key: str, mtime: Optional[float] = None, expire: Optional[timedelta] = None
    ) -> pd.DataFrame:  # pragma: no cover
        pass

    @abstractmethod
    def set(self, key: str, value: pd.DataFrame, mtime: Optional[float] = None):  # pragma: no cover
        pass

    @abstractmethod
    def delete(self, key: str):  # pragma: no cover
        pass


class InMemoryCache(Cache):
    def __init__(self):
        self._cache = {}

    def get(
        self, key: str, mtime: Optional[float] = None, expire: Optional[timedelta] = None
    ) -> pd.DataFrame:
        cached = self._cache[key]
        if self.should_invalidate(
            mtime=mtime,
            cached_mtime=cached['mtime'],
            expire=expire,
            cached_created_at=cached['created_at'],
        ):
            self.delete(key)
        return self._cache[key]['value']

    def set(self, key: str, value: pd.DataFrame, mtime: Optional[float] = None):
        mtime = mtime or time()
        self._cache[key] = {'value': value, 'mtime': mtime, 'created_at': time()}

    def delete(self, key: str):
        if key in self._cache:
            del self._cache[key]


class HDFCache(Cache):
    META_DF_KEY = '__meta__'

    def __init__(self, cache_dir: Union[str, Path]):
        self.cache_dir = Path(cache_dir).resolve()

    def get_metadata(self) -> pd.DataFrame:
        """
        metadata is a dataframe containing last mtime and created_at fields
        for each cached datasource, identified by its key (= its hash).
        If metadata file is not found or is corrupted, an empty one is recreated.
        """
        try:
            # We manually instantiate the HDFStore to be able to close it no matter what
            # See https://github.com/pandas-dev/pandas/pull/28429 for more infos
            store = pd.HDFStore(self.cache_dir / self.META_DF_KEY, mode='r')
            try:
                metadata = pd.read_hdf(store)
            finally:
                store.close()
        except Exception:  # catch all, on purpose
            metadata = pd.DataFrame(columns=['key', 'mtime', 'created_at'])
            self.set_metadata(metadata)
        return metadata

    def set_metadata(self, df: pd.DataFrame):
        df.to_hdf(self.cache_dir / self.META_DF_KEY, self.META_DF_KEY, mode='w')

    def get(
        self, key: str, mtime: Optional[float] = None, expire: Optional[timedelta] = None
    ) -> pd.DataFrame:
        metadata = self.get_metadata()
        try:
            # look for the row concerning the desired key in the metadata dataframe:
            infos = metadata[metadata.key == key].iloc[0].to_dict()
        except IndexError:
            raise KeyError(key)

        if self.should_invalidate(
            mtime=mtime,
            cached_mtime=infos['mtime'],
            expire=expire,
            cached_created_at=infos['created_at'],
        ):
            self.delete(key)

        try:
            return pd.read_hdf(self.cache_dir / key)
        except FileNotFoundError:
            raise KeyError(key)

    def set(self, key: str, value: pd.DataFrame, mtime: Optional[float] = None):
        mtime = mtime or time()
        infos = {'key': key, 'mtime': mtime, 'created_at': time()}
        metadata = self.get_metadata()
        try:
            # add new row to the metadata dataframe:
            metadata = metadata[metadata.key != key]  # drop duplicates
            metadata = metadata.append(infos, ignore_index=True)
            self.set_metadata(metadata)
            value.to_hdf(self.cache_dir / key, key, mode='w')
        except OSError:
            self.delete(key)
            raise

    def delete(self, key: str):
        metadata = self.get_metadata()
        metadata = metadata[metadata.key != key]
        self.set_metadata(metadata)
        with suppress(FileNotFoundError):
            (self.cache_dir / key).unlink()



"""
Module to add xml support
"""
from typing import Union

import pandas as pd
import pyjq
import xmltodict


def transform_with_jq(data: Union[dict, list], jq_filter: str) -> list:
    """Apply a jq filter on data before it's passed to a pd.DataFrame"""
    data = pyjq.all(jq_filter, data)

    # If the data is already presented as a list of rows,
    # then undo the nesting caused by "multiple_output" jq option
    if len(data) == 1 and (
        isinstance(data[0], list)
        or
        # detects another valid datastructure [{col1:[value, ...], col2:[value, ...]}]
        (isinstance(data[0], dict) and isinstance(list(data[0].values())[0], list))
    ):
        return data[0]  # type: ignore
    else:
        return data  # type: ignore


def read_xml(filepath: str, encoding: str = 'utf-8', filter: str = None) -> pd.DataFrame:
    data = xmltodict.parse(open(filepath).read(), encoding=encoding)
    if filter is not None:
        data = transform_with_jq(data, filter)
    return pd.DataFrame(data)



import os
from typing import List, TextIO

from ..fetcher import Fetcher, register


@register(schemes='')
class FileFetcher(Fetcher):
    def open(self, filepath) -> TextIO:
        return open(filepath)

    def listdir(self, dirpath) -> List[str]:
        return os.listdir(dirpath)

    def mtime(self, filepath) -> int:
        return int(os.path.getmtime(filepath))



import ftplib
import logging
import re
import socket
import ssl
import tempfile
from contextlib import contextmanager, suppress
from datetime import datetime
from functools import partial
from ipaddress import ip_address
from os.path import basename, join
from time import sleep
from typing import Dict, List, Optional
from typing.io import BinaryIO
from urllib.parse import ParseResult, quote, unquote, urlparse

import paramiko

FTP_SCHEMES = ['ftp', 'ftps', 'sftp']


class FTPS(ftplib.FTP_TLS):
    ssl_version = ssl.PROTOCOL_TLSv1_2

    def connect(self, host, port, timeout):
        self.host = host
        self.port = port or 990
        self.timeout = timeout

        self.sock = socket.create_connection((self.host, self.port), self.timeout)
        self.af = self.sock.family
        self.sock = self.context.wrap_socket(self.sock, server_hostname=self.host)
        self.file = self.sock.makefile('r')
        self.welcome = self.getresp()
        return self.welcome

    def ntransfercmd(self, cmd, rest=None):
        # override ntransfercmd so it reuses the sock session, to prevent SSLEOFError.
        # cf. https://stackoverflow.com/questions/40536061/ssleoferror-on-ftps-using-python-ftplib
        conn, size = ftplib.FTP.ntransfercmd(self, cmd, rest)
        if self._prot_p:
            conn = self.context.wrap_socket(
                conn, server_hostname=self.host, session=self.sock.session
            )  # this is the fix
        return conn, size

    def makepasv(self):
        # override makepasv so it rewrites the dst address if the server gave a broken one.
        # Inspired by:
        # https://github.com/lavv17/lftp/blob/d67fc14d085849a6b0418bb3e912fea2e94c18d1/src/ftpclass.cc#L774
        host, port = super().makepasv()
        if self.af == socket.AF_INET and ip_address(host).is_private:  # pragma: no cover
            host = self.sock.getpeername()[0]
        return host, port


@contextmanager
def ftps_client(params: ParseResult):
    ftps = FTPS()
    try:
        ftps.connect(host=params.hostname or '', port=params.port, timeout=3)
        ftps.prot_p()
        ftps.login(user=params.username or '', passwd=params.password or '')
        yield ftps, params.path

    finally:
        with suppress(Exception):
            ftps.quit()


@contextmanager
def ftp_client(params: ParseResult):
    port = params.port or 21
    ftp = ftplib.FTP()
    try:
        ftp.connect(host=params.hostname or '', port=port, timeout=3)
        ftp.login(user=params.username or '', passwd=params.password or '')
        yield ftp, params.path

    finally:
        with suppress(Exception):
            ftp.quit()


@contextmanager
def sftp_client(params: ParseResult):
    port = params.port or 22
    ssh_client = paramiko.SSHClient()
    try:
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(
            hostname=params.hostname,  # type: ignore
            username=params.username,
            password=params.password,
            port=port,
            timeout=3,
        )
        sftp = ssh_client.open_sftp()
        yield sftp, params.path

    finally:
        ssh_client.close()


def _urlparse(url: str) -> ParseResult:
    """
    We had several cases of ftp usernames having '#' in them which confuses urlparse.
    The # is a reserved char for 'fragment identifiers':
    https://en.wikipedia.org/wiki/Fragment_identifier
    """
    url_params = urlparse(quote(url, safe=':/@'))  # the '#' is marked as unsafe
    # unquote to get back the original bits of string
    return ParseResult(*[unquote(param) for param in url_params])


def client(url):
    parse_result = _urlparse(url)
    ftp_client_mapping = {'ftp': ftp_client, 'ftps': ftps_client, 'sftp': sftp_client}
    return ftp_client_mapping[parse_result.scheme](parse_result)


def retry_pasv(c, cmd, *args):
    """
    Some servers accept the PASV command, but timeout when doing commands using it.
    Solution is to retry a timeout-ed command in active mode.
    """
    fun = partial(getattr(c, cmd), *args)
    try:
        return fun()
    except socket.timeout:
        c.set_pasv(False)
        return fun()


def _open(url: str) -> BinaryIO:
    ret = tempfile.NamedTemporaryFile(suffix='.ftptmp')
    with client(url) as (c, path):
        try:
            retry_pasv(c, 'retrbinary', f'RETR {path}', ret.write)
        except AttributeError:
            c.getfo(path, ret)
        except ftplib.error_perm as e:
            raise Exception(f"Can't open file {path}. Please make sure the file exists") from e

    ret.seek(0)
    return ret


def ftp_open(url: str, retry: int = 4) -> BinaryIO:
    for i in range(1, retry + 1):
        try:
            return _open(url)
        except (AttributeError, OSError, ftplib.error_temp) as e:
            sleep_time = 2 * i ** 2
            logging.getLogger(__name__).warning(f'Retry #{i}: Sleeping {sleep_time}s because {e}')
            sleep(sleep_time)


def _get_all_files(c, path) -> List[str]:
    try:
        # retry_pasv returns path + the file
        return [basename(x) for x in retry_pasv(c, 'nlst', path)]
    except AttributeError:
        if not path:
            path = '.'
        return c.listdir(path)


def ftp_listdir(url: str) -> List[str]:
    with client(url) as (c, path):
        return _get_all_files(c, path)


def _get_mtime(c, path) -> Optional[int]:
    """Returns timestamp of last modification"""
    try:
        mdtm = c.sendcmd('MDTM ' + path)
        # mdtm-response = "213" SP time-val CRLF (e.g. '20180101203000')
        # some FTP servers response include the milliseconds (e.g. '20180101203000.123')
        mdtm = re.search(r'^213 (\d+)(\.\d+)?$', mdtm).group(1)  # type: ignore
        dt = datetime.strptime(mdtm, '%Y%m%d%H%M%S')
        return int((dt - datetime(1970, 1, 1)).total_seconds())
    except AttributeError:
        return c.stat(path).st_mtime
    except ftplib.error_perm as e:
        logging.getLogger(__name__).warning(
            f"Can't open file {path}. Please make sure the file exists: {e}"
        )
        return None


def ftp_mtime(url: str) -> Optional[int]:
    with client(url) as (cl_ftp, path):
        return _get_mtime(cl_ftp, path)


def dir_mtimes(url: str) -> Dict[str, Optional[int]]:
    with client(url) as (cl_ftp, path):
        mtimes_dir = {}
        all_files = _get_all_files(cl_ftp, path)
        for file in all_files:
            mtimes_dir[file] = _get_mtime(cl_ftp, join(path, file))
    return mtimes_dir



from os import path
from typing import Dict, Hashable

from .cache import Cache
from .datasource import DataSource


class DataPool:
    def __init__(
        self, config: Dict[Hashable, dict], data_sources_dir: str = '', cache: Cache = None
    ):
        self.cache = cache
        self.datasources: dict = {}
        for ds_id, ds_conf in config.items():
            ds = DataSource(**ds_conf)

            # change local path into absolute path
            if ds.scheme == '' and not path.isabs(ds.uri):
                ds.uri = path.join(data_sources_dir, ds.uri)

            self.datasources[ds_id] = ds

    def __contains__(self, item):
        return item in self.datasources

    def __getitem__(self, item):
        return self.datasources[item].get_df(cache=self.cache)

    def __len__(self):
        return len(self.datasources)



import tempfile
from email.utils import parsedate_to_datetime
from typing import IO, List, Optional

import certifi
import urllib3

from ..fetcher import Fetcher, register


@register(schemes=['http', 'https'])
class HttpFetcher(Fetcher):
    def __init__(self, *, verify=True, **kwargs):
        super().__init__(**kwargs)
        if verify:
            self.pool_manager = urllib3.PoolManager(
                cert_reqs='CERT_REQUIRED', ca_certs=certifi.where()
            )
        else:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            self.pool_manager = urllib3.PoolManager(cert_reqs='CERT_NONE', assert_hostname=False)

    def open(self, filepath) -> IO:
        r = self.pool_manager.request('GET', filepath, preload_content=False, **self.extra_kwargs)
        ret = tempfile.NamedTemporaryFile(suffix='.httptmp')
        for chunk in r.stream():
            ret.write(chunk)
        ret.seek(0)
        return ret

    def listdir(self, dirpath) -> List[str]:
        raise NotImplementedError

    def mtime(self, filepath) -> Optional[int]:
        try:
            r = self.pool_manager.request('HEAD', filepath)
        except Exception:
            return None
        else:
            dt = parsedate_to_datetime(r.headers['last-modified'])
            return int(dt.timestamp())



# flake8: noqa
from .fetcher import Fetcher, MatchEnum
from .ftp.ftp_fetcher import FTPFetcher
from .http.http_fetcher import HttpFetcher
from .local.file_fetcher import FileFetcher
from .s3.s3_fetcher import S3Fetcher



"""
Module to enhance pandas.read_json with JQ filter
"""
import json
from functools import wraps

import pandas as pd
import pyjq


def transform_with_jq(json_input: str, jq_filter: str) -> str:
    """Apply a jq filter on raw json input, outputs raw json (may be on several lines)"""
    return '\n'.join(json.dumps(x) for x in pyjq.all(jq_filter, json.loads(json_input)))


@wraps(pd.read_json)
def read_json(
    path_or_buf=None, encoding: str = 'utf-8', filter: str = None, *args, **kwargs
) -> pd.DataFrame:
    if filter is not None:
        with open(path_or_buf, encoding=encoding) as f:
            path_or_buf = transform_with_jq(f.read(), filter)
    return pd.read_json(path_or_buf, encoding=encoding, *args, **kwargs)
